/*
 * Copyright (c) 2015 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.util.Random;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Generates unique, sequential IDs
 * 
 * @author F. Mannhardt
 * 
 */
final class IdFactoryRandomPermutation implements IdFactory {

	private static final long PRIME = 4294967291l;
	private static final long HALFPRIME = PRIME / 2;

	private long seed = permuteQPR(new Random().nextInt() ^ 0x46790905);

	private final int intervalShift;
	private AtomicLong index;

	/**
	 * Creates a new ID factory.
	 */
	IdFactoryRandomPermutation(int intervalShift, Long initialValue) {
		this.intervalShift = intervalShift;
		if (initialValue != null) {
			this.index = new AtomicLong(initialValue / intervalShift);
		} else {
			this.index = new AtomicLong();
		}
	}

	private static long permuteQPR(long x) {
		if (x >= PRIME) {
			// Give up, ran out of numbers the following ones are just sequential
			return x;
		} else if (x <= HALFPRIME) {
			return quadraticResidue(x);
		} else {
			return PRIME - quadraticResidue(x);
		}
	}

	private static long quadraticResidue(long x) {
		return (x * x) % PRIME;
	}

	/**
	 * Creates a new, unique ID.
	 * 
	 * @return Unique ID.
	 */
	public synchronized long nextId() {
		long currentIndex = index.incrementAndGet();
		long permutatedIndex = permuteQPR(permuteQPR(currentIndex) + seed ^ 0x5bf03635l);
		return permutatedIndex << intervalShift;
	}
	
	public int getIntervalCapacity() {
		return 1 << intervalShift;
	}
	
	public int getIdShift() {
		return intervalShift;
	}

}